#coding: utf-8

from __future__ import print_function

import licant
from licant.core import Target, UpdatableTarget, UpdateStatus, core
from licant.cache import fcache
from licant.util import red, green, yellow, purple, quite
import threading
from licant.util import deprecated
import os
import sys

_rlock = threading.RLock()


def do_execute(target, rule, msgfield, prefix=None):
	def sprint(*args, **kwargs):
		with _rlock:
			print(*args, **kwargs)

	rule = rule.format(**target.__dict__)

	message = getattr(target, msgfield, None)

	if not core.runtime["quite"]:
		if not core.runtime["debug"] and message is not None:
			if not isinstance(message, quite):
				if prefix is not None:
					sprint(prefix, message.format(**target.__dict__))
				else:
					sprint(message.format(**target.__dict__))
	
		else:
			sprint(rule)

	ret = os.system(rule)

	if target.isfile == True:
		target.update_info(target)

	return True if ret == 0 else False


class Executor:
	def __init__(self, rule, msgfield='message'):
		self.rule = rule
		self.msgfield = msgfield

	def __call__(self, target, **kwargs):
		return do_execute(target, self.rule, self.msgfield, **kwargs)

class MakeFileTarget(UpdatableTarget):
	__actions__ = UpdatableTarget.__actions__.union({"clean", "makefile"})

	def __init__(self, tgt, deps, **kwargs):
		UpdatableTarget.__init__(self, tgt, deps, **kwargs)

	def clean(self):
		stree = self.core.subtree(self.tgt)
		return stree.invoke_foreach(ops="clr", cond=if_file_and_exist)

	def makefile(self):
		return self.recurse_update()
		#stree = subtree(self.tgt)
		#stree.invoke_foreach(ops="dirkeep")
		#stree.reverse_recurse_invoke(
		#    ops="update_if_need", threads=core.runtime["threads"])


class FileTarget(MakeFileTarget):
	__actions__ = MakeFileTarget.__actions__.union({"build", "clr"})

	def __init__(self, tgt, deps, force=False, **kwargs):
		MakeFileTarget.__init__(self, tgt, deps, **kwargs)
		self.isfile = True
		self.force = force
		self.clrmsg = "DELETE {tgt}"
		self.default_action = "makefile"

	def update_info(self, _self):
		fcache.update_info(self.tgt)
		return True

	def mtime(self):
		curinfo = fcache.get_info(self.tgt)
		if curinfo.exist == False:
			return 0
		else:
			return curinfo.mtime

	def dirkeep(self):
		"""Create directory tree for this file if needed."""
		dr = os.path.normpath(os.path.dirname(self.tgt))
		if (not os.path.exists(dr)):
			print("MKDIR %s" % dr)
			os.system("mkdir -p {0}".format(dr))
		return True

	def is_exist(self):
		curinfo = fcache.get_info(self.tgt)
		return curinfo.exist

	def warn_if_not_exist(self):
		"""Print warn if file isn't exist"""
		info = fcache.get_info(self.tgt)
		if info.exist == False:
			print("Warn: file {} isn`t exist".format(purple(self.tgt)))

	def clr(self):
		"""Delete this file."""
		do_execute(self, "rm -f {tgt}", "clrmsg")

	def self_need(self):
		if self.force or self.is_exist() == False:
			return True

		maxtime = 0
		for dep in self.get_deplist():
			if dep.mtime() > maxtime:
				maxtime = dep.mtime()

		if maxtime > self.mtime():
			return True

		return False

	def update(self):
		self.dirkeep()
		return self.build(self)


class FileSet(MakeFileTarget):
	"""Virtual file target`s set.

	For link a set of file objects to the licant tree
	without depend`s overhead."""

	def __init__(self, tgt, targets):
		MakeFileTarget.__init__(
			self, tgt=tgt, deps=targets, default_action="makefile")
		self.targets = targets
		self.__mtime = None

	def self_need(self):
		return False

	def update(self):
		pass

	def mtime(self):
		if (self.__mtime is None):
			maxtime = 0
			for dep in self.get_deplist():
				if dep.mtime() > maxtime:
					maxtime = dep.mtime()
			self.__mtime = maxtime

		return self.__mtime


def source(tgt, deps=[]):
	"""Index source file by licant core."""
	target = FileTarget(
		build=lambda self: self.warn_if_not_exist(),
		deps=deps,
		tgt=tgt,
	)
	target.clr = None
	target.dirkeep = licant.util.do_nothing
	target.update_status = UpdateStatus.Keeped
	core.add(target)


def copy(tgt, src, adddeps=[], message="COPY {src} {tgt}"):
	"""Make the file copy target."""
	core.add(FileTarget(
		tgt=tgt,
		build=Executor("cp {src} {tgt}"),
		src=src,
		deps=[src] + adddeps,
		message=message
	))


def fileset(tgt, targets):
	"""Make a fileset."""
	core.add(FileSet(
		tgt=tgt,
		targets=targets
	))


def if_file_and_exist(target):
	if not isinstance(target, FileTarget):
		return False

	curinfo = fcache.get_info(target.tgt)
	return curinfo.exist


#coding: utf-8

from licant.modules import mlibrary
from licant.cxx_make import host_binutils, binutils
from licant.util import red, yellow, cxx_read_depends

import os
import licant.util as gu
import licant.make

import glob


class solver:
    def __init__(self, type, merge, include, default):
        self.type = type
        self.merge = merge
        self.include = include
        self.default = default


def base(base, local, solver, tbase, tlocal):
    return base


def local(base, local, solver, tbase, tlocal):
    if local is None:
        return solver.default
    return local


def concat(base, local, solver, tbase, tlocal):
    if local is None:
        local = []
    if base is None:
        base = []
    return base + local


def strconcat(base, local, solver, tbase, tlocal):
    if local is None:
        return base
    return base + " " + local


def concat_add_srcdir(base, local, solver, tbase, tlocal):
    if local is None:
        local = []
    else:
        if "srcdir" in tlocal.opts:
            srcdir = tlocal.opts["srcdir"]
        else:
            srcdir = ""

        local = list(map(lambda p: os.path.join(
            tlocal.opts["__dir__"], srcdir, os.path.expanduser(p)), local))
    #print(base + local)
    return base + local


def local_add_srcdir(base, local, solver, tbase, tlocal):
    if "srcdir" in tlocal.opts:
        srcdir = tlocal.opts["srcdir"]
    else:
        srcdir = ""
    if isinstance(local, type("str")):
        local = [local]
    if local is None:
        return []
    return list(map(lambda p: os.path.join(tlocal.opts["__dir__"], srcdir, os.path.expanduser(p)), local))


def concat_add_locdir(base, local, solver, tbase, tlocal):
    if local is None:
        return base

    if isinstance(local, type("str")):
        local = [local]
    local = list(map(lambda p: os.path.join(
        tlocal.opts["__dir__"], os.path.expanduser(p)), local))
    return base + local


def concat_add_locdir_second(base, local, solver, tbase, tlocal):
    """Добавить локальный путь ко всем вторым элементам локального массива. 
    Для поддержки локальных заголовков."""
    if local is None:
        return base
    local = [(pl[0], os.path.join(tlocal.opts["__dir__"], pl[1]))
             for pl in local]
    return base + local


def local_if_exist(base, local, solver, tbase, tlocal):
    return base if local is None else local


def concat_to_submodule(base, local, solver, tbase, tlocal):
    if local is None:
        local = []
    if base is None:
        base = []
    lm = [licant.modules.submodule(l) for l in local]
    return base + lm


cxx_module_field_list = {
    # merge						#include
    "srcdir": 			solver("str", 		local, 						base,					"."),
    "objects": 			solver("list", 		local_add_srcdir,			concat_add_srcdir,		[]),
    "sources": 			solver("list", 		local_add_srcdir,			concat_add_srcdir,		[]),
    "moc": 				solver("list", 		local_add_srcdir,			concat_add_srcdir,		[]),
    "libs": 			solver("list", 		concat,						concat,					[]),
    "target": 			solver("str", 		local_if_exist,				base,					"target"),
    "include_paths": 	solver("list", 		concat_add_locdir, 			concat_add_locdir,		[]),
    "cxxstd": 			solver("str", 		local_if_exist, 			local_if_exist,			"c++14"),
    "ccstd": 			solver("str", 		local_if_exist, 			local_if_exist,			"c11"),
    "cxx_flags": 		solver("str", 		strconcat, 					strconcat,				""),
    "cc_flags": 		solver("str", 		strconcat, 					strconcat,				""),
    "ld_flags": 		solver("str", 		strconcat, 					strconcat,				""),
    "modules": 			solver("list", 		local, 						concat,					[]),
    "type": 			solver("str", 		local,			 			base,					"objects"),
    "builddir": 		solver("str", 		local_if_exist, 			base,					"build"),
    "binutils": 		solver("binutils", 	local_if_exist, 			base,					host_binutils),
    "include_modules": 	solver("list", 		concat_to_submodule,		base,					[]),
    "defines": 			solver("list", 		concat, 					concat,					[]),
    "ldscripts":		solver("list", 		concat_add_locdir, 			concat_add_locdir,		[]),
    "local_headers":	solver("list", 		concat_add_locdir_second, 	concat_add_locdir_second,	[]),
}


class CXXModuleOptions:
    def __getitem__(self, key):
        return self.opts[key]

    def check(self):
        for k, v in self.opts.items():
            if not k in licant.modules.special:
                if not k in self.table:
                    print("Unresolved option: {}".format(red(k)))
                    exit(-1)

                if v.__class__.__name__ != self.table[k].type:
                    print("Option should be object of {}: {}".format(
                        yellow(self.table[k].type), red(k)))
                    exit(-1)

    def set_default_if_empty(self, table=cxx_module_field_list):
        for k in table:
            if not k in self.opts:
                self.opts[k] = table[k].default

    def merge(self, other, fnc):
        resopts = {}
        for k, v in self.opts.items():

            base = self.opts[k]
            if k in other.opts:
                local = other.opts[k]
            else:
                local = None

            func = getattr(self.table[k], fnc)
            resopts[k] = func(base, local, self.table[k], self, other)

        return CXXModuleOptions(**resopts)

    def __init__(self, table=cxx_module_field_list, **kwargs):
        self.opts = kwargs
        self.table = table
        self.check()


def cxx_options_from_modopts(modopts):
    cxxstd = "-std=" + modopts["cxxstd"]
    ccstd = "-std=" + modopts["ccstd"]

    cxx_flags = cxxstd + " " + modopts["cxx_flags"]
    cc_flags = ccstd + " " + modopts["cc_flags"]

    ld_srcs_add = "".join([" -l" + l for l in modopts["libs"]])

    return licant.cxx_make.options(
        binutils=modopts["binutils"],
        include_paths=modopts["include_paths"],
        cc_flags=cc_flags,
        cxx_flags=cxx_flags,
        defines=modopts["defines"],
        ldscripts=modopts["ldscripts"],
        ld_flags=modopts["ld_flags"],
        ld_srcs_add=ld_srcs_add
    )


def build_paths(srcs, opts, ext):
    objs = []
    for s in srcs:
        if os.path.isabs(s):
            objs.append(opts.opts["builddir"] + gu.changeext(s, ext))
        else:
            objs.append(os.path.normpath(os.path.join(
                opts.opts["builddir"], gu.changeext(s, ext).replace("..", "__"))))
    return objs


def sources_paths(opts):
    ret = []
    for s in opts["sources"]:
        if "*" in s:
            ret.extend(glob.glob(s))
        else:
            ret.append(s)
    return ret


def moc_paths(opts):
    ret = []
    for s in opts["moc"]:
        if "*" in s:
            ret.extend(glob.glob(s))
        else:
            ret.append(s)
    return ret


def link_objects(srcs, objs, deps, cxxopts, adddeps):

    for s, o, d in zip(srcs, objs, deps):
        if not s in licant.core.core.targets:
            licant.make.source(s)

        headers = cxx_read_depends(d)
        force = headers is None
        if headers is None:
            headers = []
        else:
            for h in headers:
                if h not in licant.core.core.targets:
                    licant.make.source(h)

        licant.cxx_make.depend(src=s, tgt=d, opts=cxxopts, deps=[
                               s] + adddeps + headers, force=force)
        licant.cxx_make.object(src=s, tgt=o, opts=cxxopts, deps=[
                               s, d] + adddeps + headers)


def link_moc(mocs, srcs, cxxopts, adddeps):
    for m, s in zip(mocs, srcs):
        licant.make.source(m)
        licant.cxx_make.moc(src=m, tgt=s, opts=cxxopts, deps=[m] + adddeps)


def executable(srcs, opts):
    cxxopts = cxx_options_from_modopts(opts)
    licant.cxx_make.executable(tgt=opts["target"], srcs=srcs, opts=cxxopts)
    return opts["target"]


def dynlib(srcs, opts):
    cxxopts = cxx_options_from_modopts(opts)
    licant.cxx_make.dynamic_library(
        tgt=opts["target"], srcs=srcs, opts=cxxopts)
    return opts["target"]


def virtual(srcs, opts):
    licant.core.target(tgt=opts["target"], deps=srcs)
    return opts["target"]


def make(name, impl=None, **kwargs):
    opts = CXXModuleOptions(**kwargs)
    opts.set_default_if_empty()

    def modmake(name, impl, baseopts):
        mod = mlibrary.get(name, impl)

        moddir = os.path.dirname(mod.script)

        modopts = CXXModuleOptions(**mod.opts)
        locopts = baseopts.merge(modopts, "merge")

        def include_modules(locopts, lst):
            retopts = locopts
            for simod in lst:
                imod = mlibrary.get(simod.name, simod.impl)
                retopts = retopts.merge(imod, "include")

                if "include_modules" in imod.opts:
                    retopts = include_modules(
                        retopts, imod.opts["include_modules"])

            return retopts

        adddeps = []  # mod.stack

        locopts = include_modules(locopts, locopts["include_modules"])

        local_headers_targets = []
        if len(locopts["local_headers"]):
            locopts["include_paths"].append(
                locopts["builddir"] + "/__LOCAL_HEADERS__/")
            for pair in locopts["local_headers"]:
                licant.make.source(pair[1])
                tgtpath = os.path.join(
                    locopts["builddir"], "__LOCAL_HEADERS__", pair[0])
                licant.make.copy(src=pair[1], tgt=tgtpath)

                local_headers_targets.append(tgtpath)

        licant.make.fileset(name + "__local_headers__", local_headers_targets)
        adddeps.append(name + "__local_headers__")

        cxxopts = cxx_options_from_modopts(locopts)

        locsrcs = sources_paths(locopts)
        locobjs = build_paths(locsrcs, locopts, "o")
        locdeps = build_paths(locsrcs, locopts, "d")

        if len(locopts["moc"]) != 0:
            locmoc = moc_paths(locopts)
            locmoccxx = build_paths(locmoc, locopts, "h.cxx")
            locmocobjs = build_paths(locmoc, locopts, "h.cxx.o")
            locmocdeps = build_paths(locmoc, locopts, "h.cxx.d")
            link_moc(locmoc, locmoccxx, cxxopts, adddeps)

            locsrcs.extend(locmoccxx)
            locobjs.extend(locmocobjs)
            locdeps.extend(locmocdeps)

        link_objects(locsrcs, locobjs, locdeps, cxxopts, adddeps)

        if len(locopts["objects"]) != 0:
            for t in locopts["objects"]:
                licant.make.source(t)
            locobjs.extend(locopts["objects"])

        submodules_results = []
        # print(locopts["modules"])
        for smod in locopts["modules"]:
            submodules_results += modmake(smod.name, smod.impl, locopts)

        if locopts["type"] == "application":
            return executable(locobjs + submodules_results, locopts)
        if locopts["type"] == "shared_library":
            return dynlib(locobjs + submodules_results, locopts)
        elif locopts["type"] == "objects":
            return locobjs + submodules_results
        else:
            print("Wrong type of assemble: {}", gu.red(locopts["type"]))
            exit(-1)

    res = modmake(name, impl, opts)
    return res


def task(name, target, impl, type, **kwargs):
    if target is None:
        target = name
    else:
        licant.make.fileset(tgt=name, targets=[target])
    licant.modules.module(name, impl=impl, type=type, target=target, **kwargs)
    make(name)


def application(name, target=None, impl=None, type="application", **kwargs):
    return task(name, target, impl, type, **kwargs)


def shared_library(name, target=None, impl=None, type="shared_library", **kwargs):
    return task(name, target, impl, type, **kwargs)



from licant.scripter import scriptq
from licant.util import yellow
import licant.core

import os
import sys
import json

gpath = "/var/lib/licant"
lpath = os.path.expanduser("~/.licant")

libs = None


def merge_two_dicts(x, y):
	z = x.copy()   # start with x's keys and values
	z.update(y)    # modifies z with y's keys and values & returns None
	return z


def init():
	global libs

	glibs = {}
	llibs = {}

	if os.path.exists(gpath):
		glibs = json.load(open(gpath))

	if os.path.exists(lpath):
		llibs = json.load(open(lpath))

	libs = merge_two_dicts(glibs, llibs)

def include(lib):
	if libs is None:
		init()

	if not lib in libs:
		print("Unregistred library {}. Use licant-config utility or manually edit {} or {} file.".format(
			yellow(lib), yellow(lpath), yellow(gpath)))
		exit(-1)

	scriptq.execute(libs[lib])

def print_libs(taget, *args):
	if libs is None:
		init()
		
	keys = sorted(libs.keys())
	for k in keys:
		print("{}: {}".format(k,libs[k]))


libs_target = licant.core.Target(
	tgt="l",
	deps=[],
	list=print_libs,
	actions={"list"}
)

licant.core.core.add(libs_target)



# coding: utf-8

import licant.util
import threading
import types
import inspect
from enum import Enum
import re

class WrongAction(Exception):
	pass


class NoneDictionary(dict):
	def __init__(self): dict.__init__(self)

	def __getitem__(self, idx):
		try:
			return dict.__getitem__(self, idx)
		except:
			return None


class Core:
	def __init__(self):
		self.targets = {}
		self.runtime = NoneDictionary()

	def add(self, target):
		"""Add new target"""
		target.core = self
		self.targets[target.tgt] = target

	def get(self, tgt):
		"""Get target object"""
		if tgt in self.targets:
			return self.targets[tgt]
		licant.util.error("unregistred target " + licant.util.yellow(tgt))

	def subtree(self, root):
		"""Construct Subtree accessor for root target"""
		return SubTree(self, root)

	def depends_as_set(self, tgt, incroot=True):
		""""""
		res = set()
		if incroot:
			res.add(tgt)

		target = self.get(tgt)

		for d in target.deps:
			if not d in res:
				res.add(d)
				subres = self.depends_as_set(d)
				res = res.union(subres)
		return sorted(res)


# Объект ядра с которым библиотеки работают по умолчанию.
core = Core()


class SubTree:
	def __init__(self, core, root):
		self.root = root
		self.core = core
		self.depset = core.depends_as_set(root)

#    def update(self):
#        SubTree.__init__(self, root)

	def invoke_foreach(self, ops, cond=licant.util.always_true):
		if core.runtime["debug"]:
			print("FOREACH(root={}, ops={}, cond={})".format(self.root, ops, cond))

		sum = 0
		ret = None

		for d in self.depset:
			target = self.core.get(d)
			if cond(target):
				ret = target.invoke(ops)

			if ret is not None:
				sum += 1

		return sum

	def __generate_rdepends_lists(self, targets):
		for t in targets:
			t.rdepends = []
			t.rcounter = 0

		for t in targets:
			for dname in t.deps:
				dtarget = self.core.get(dname)
				dtarget.rdepends.append(t.tgt)

	def reverse_recurse_invoke_single(self, ops, threads=None, cond=licant.util.always_true):
		if core.runtime["trace"]:
			print("SINGLETHREAD MODE(root={}, ops={}, cond={})".format(
				self.root, ops, cond))
		targets = [self.core.get(t) for t in self.depset]

		self.__generate_rdepends_lists(targets)

		works = licant.util.queue()

		for t in targets:
			if t.rcounter == len(t.deps):
				works.put(t)

		while(not works.empty()):
			w = works.get()

			if cond(w):
				ret = w.invoke(ops)
				if ret is False:
					print(licant.util.red("runtime error"))
					exit(-1)

			for r in [self.core.get(t) for t in w.rdepends]:
				r.rcounter = r.rcounter + 1
				if r.rcounter == len(r.deps):
					works.put(r)

	def reverse_recurse_invoke_threads(
			self, ops, threads,
			cond=licant.util.always_true
	):
		if core.runtime["trace"]:
			print("MULTITHREAD MODE(root={}, threads={}, ops={}, cond={})"
				  .format(self.root, threads, ops, cond))

		targets = [self.core.get(t) for t in self.depset]

		self.__generate_rdepends_lists(targets)
		works = licant.util.queue()

		class info_cls:
			def __init__(self):
				self.have_done = 0
				self.need_done = len(targets)
				self.sum = 0
				self.err = False
		info = info_cls()

		for t in targets:
			if t.rcounter == len(t.deps):
				works.put(t)

		lock = threading.Lock()

		def thread_func(index):
			while info.have_done != info.need_done:
				if info.err:
					break

				lock.acquire()
				if not works.empty():
					w = works.get()
					lock.release()

					if core.runtime["trace"]:
						print(
							"TRACE: THREAD {0} get work {1}".format(index, w))

					if cond(w):
						try:
							ret = w.invoke(ops)
						except:
							info.err = True
							return
						if ret is False:
							info.err = True
							return
						if ret == 0:
							info.sum += 1

					for r in [self.core.get(t) for t in w.rdepends]:
						r.rcounter = r.rcounter + 1
						if r.rcounter == len(r.deps):
							works.put(r)

					info.have_done += 1

					if core.runtime["trace"]:
						print(
							"TRACE: THREAD {0} finished with work {1}".format(index, w))

					continue
				lock.release()

		threads_list = [
			threading.Thread(target=thread_func, args=(i,))
			for i in range(0, threads)
		]

		for t in threads_list:
			t.start()

		for t in threads_list:
			t.join()

		if info.err:
			print(licant.util.red("runtime error (multithreads mode)"))
			exit(-1)
		return info.sum

	def reverse_recurse_invoke(self, *args, **kwargs):
		if "threads" in kwargs:
			if kwargs["threads"] == 1:
				return self.reverse_recurse_invoke_single(*args, **kwargs)
			else:
				return self.reverse_recurse_invoke_threads(*args, **kwargs)
		else:
			return self.reverse_recurse_invoke_single(*args, **kwargs)

	def __str__(self):
		ret = ""
		for d in sorted(self.depset):
			t = self.core.get(d)
			s = "{}: {}\n".format(d, sorted(t.deps))
			ret += s
		ret = ret[:-1]
		return ret


class Target:
	__actions__ = {"actlist"}

	def __init__(self, tgt, deps, actions=None, **kwargs):
		self.tgt = tgt
		self.deps = set(deps)
		for k, v in kwargs.items():
			setattr(self, k, v)

		if actions is not None:
			self.__actions__ = self.__actions__.union(set(actions))

	def get_deplist(self):
		return [self.core.get(d) for d in self.deps]

	def actlist(self):
		print(licant.util.get_actions(self))

	def invoke(self, funcname, *args, critical=False, **kwargs):
		"""Invoke func function or method, or mthod with func name for this target

		Поддерживается несколько разных типов func.
		В качестве func может быть вызвана внешняя функция с параметром текущей цели,
		или название локального метода.
		critical -- Действует для строкового вызова. Если данный attr отсутствует у цели,
		то в зависимости от данного параметра может быть возвращен None или выброшено исключение.
		"""
		if core.runtime["trace"]:
			print("TRACE: Invoke: tgt:{}, act:{}, args:{}, kwargs:{}".format(self.tgt, funcname, args, kwargs))

		if funcname not in self.__actions__:
			licant.error("Isn't action {}".format(licant.util.yellow(funcname)))

		func = getattr(self, funcname, None)
		if (func is None):
			if (critical):
				raise WrongAction(func)
			return None
		
		if isinstance(func, types.MethodType):
			#return licant.util.cutinvoke(func, *args, **kwargs)
			return func(*args, **kwargs)

		else:
			#return licant.util.cutinvoke(func, self, *args, **kwargs)
			return func(self, *args, **kwargs)

	def __repr__(self):
		"""По умолчанию вывод Target на печать возвращает идентификатор цели"""
		return self.tgt


class UpdateStatus(Enum):
	Waiting = 0
	Keeped = 1
	Updated = 2


class UpdatableTarget(Target):
	__actions__ = Target.__actions__.union({"recurse_update", "update", "update_if_need"})

	def __init__(
			self, tgt, deps, default_action="recurse_update",
			update_status=UpdateStatus.Waiting, **kwargs
	):
		Target.__init__(self, tgt, deps,
						default_action=default_action, **kwargs)
		self.update_status = update_status

	def update(self):
		licant.error("Unoverrided update method")

	def self_need(self):
		return False

	def has_updated_depends(self):
		depends = self.get_deplist()

		for d in depends:
			if not isinstance(d, UpdatableTarget) or d.update_status == UpdateStatus.Updated:
				return True

			if d.update_status == UpdateStatus.Waiting:
				print(d)
				licant.error("Unwalked depends in UpdatableTarget")

		return False

	def update_if_need(self):
		if self.has_updated_depends() or self.self_need(): #self.invoke("self_need"):
			self.update_status = UpdateStatus.Updated
			return self.update() #self.invoke("update")
		else:
			self.update_status = UpdateStatus.Keeped
			return True

	def recurse_update(self):
		stree = self.core.subtree(self.tgt)
		stree.reverse_recurse_invoke(
			ops="update_if_need", threads=core.runtime["threads"])


class Routine(UpdatableTarget):
	def __init__(self, func, deps=[], tgt=None, **kwargs):
		if tgt is None:
			tgt = func.__name__
		UpdatableTarget.__init__(self, tgt=tgt, deps=deps, **kwargs)
		self.func = func

	def update(self):
		return self.func()

	def self_need(self):
		return True

def routine_decorator(func=None, deps=None):
	if inspect.isfunction(func):
		core.add(Routine(func))
		return func

	else:
		def decorator(func):
			core.add(Routine(func, deps=deps))
			return func
			
		return decorator

def print_targets_list(target, *args):
	if core.runtime["debug"]:
		print("print_targets_list args:{}".format(args))

	if len(core.targets) == 0:
		print("targets doesn't founded")
		return

	keys = sorted(core.targets.keys())

	if len(args) > 0:
		keys = [m for m in keys if re.search(args[0], m)]

	for k in keys:
		v = core.targets[k]
		print(k)

def print_target_info(taget, *args):
	if len(args) == 0:
		licant.error("Need target mnemo")

	print("name:", core.get(args[0]))
	print("deps:", sorted(core.get(args[0]).deps))



corediag_target = Target(
	tgt="corediag",
	deps=[],
	targets=print_targets_list,
	tgtinfo=print_target_info,
	actions={"targets", "tgtinfo"}
)

core.add(corediag_target)

